"""
    Scraper to get data from El Pitazo
"""
# Internal imports
from c4v.scraper.scrapers.base_scrapy_scraper import BaseScrapyScraper
from c4v.scraper.spiders.el_pitazo import ElPitazoSpider


class ElPitazoScraper(BaseScrapyScraper):
    """
        Scrapes data from ElPitazo, relies in 
        scrapy for this.
    """

    intended_domain = "elpitazo.net"
    spider = ElPitazoSpider



"""
    This file implements a class that manages local storage 
    using regular files.
"""
# Python
from sqlite3.dbapi2 import connect
import sqlite3
from typing import Any, Iterator, Dict, List
import dataclasses
import datetime

# Local imports
from c4v.scraper.persistency_manager.base_persistency_manager import (
    BasePersistencyManager,
)
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData
from config import settings

DATE_FORMAT = settings.date_format

class SqliteManager(BasePersistencyManager):
    """
        Use SQLite based local storage 
        to manage persistency.
    """

    def __init__(self, db_path: str):
        self._db_path = db_path
        self._migrate(db_path)

    def _migrate(self, db_path: str):
        """
            Perform a database migration, ensuring that tables are as specified in this class,
            use db specified by given path
        """
        with sqlite3.connect(db_path) as connection:
            cursor = connection.cursor()  # connect to cursor

            # All these migations are "if not exists" so no change is performed when they already exist
            self._create_scraped_data_table(cursor)  # create scraped data table
            self._create_category_table(
                cursor
            )  # category table (including many to many)
            connection.commit()  # save changes

    def _create_scraped_data_table(self, cursor: sqlite3.Cursor):
        """
            Create scraped data with the structure as in ScrapedData model, perform changes with 
            provided cursor
        """
        cursor.execute(
            "CREATE TABLE IF NOT EXISTS  scraped_data(\
            url PRIMARY KEY,\
            last_scraped DATETIME NULL,\
            title TEXT NULL,\
            content TEXT NULL,\
            author TEXT NULL,\
            date DATETIME NULL\
            );"
        )

    def _create_category_table(self, cursor: sqlite3.Cursor):
        """
            Create category table, including a many to many relationship between categories.
            Perform changes with provided cursor
        """
        cursor.execute(
            "CREATE TABLE IF NOT EXISTS category(\
            name PRIMARY KEY\
        );"
        )
        cursor.execute(
            "CREATE TABLE IF NOT EXISTS  category_to_data (\
                scraped_data_url,\
                category_name,\
                FOREIGN KEY (scraped_data_url) REFERENCES scraped_data(url) ON DELETE CASCADE,\
                FOREIGN KEY (category_name) REFERENCES category(name) ON DELETE CASCADE,\
                CONSTRAINT uniq UNIQUE (scraped_data_url, category_name)\
            );"
        )

    def get_all(self, limit : int, scraped : bool) -> Iterator[ScrapedData]:

        # Retrieve all data stored
        with sqlite3.connect(self._db_path) as connection:
            cursor = connection.cursor()
            if scraped:
                new_cur = cursor.execute("SELECT * FROM scraped_data WHERE last_scraped IS NOT NULL;",)
            elif scraped==False:
                new_cur = cursor.execute("SELECT * FROM scraped_data WHERE last_scraped IS NULL;",)
            else:
                new_cur = cursor.execute("SELECT * FROM scraped_data;",)

            if limit < 0:
                res = new_cur.fetchall()
            else:
                res = new_cur.fetchmany(limit)

            for row in res:
                # Decompose row
                (url, last_scraped, title, content, author, date) = row

                # parse date to datetime:
                last_scraped = datetime.datetime.strptime(last_scraped, DATE_FORMAT) if last_scraped else last_scraped

                categories = [
                    category
                    for (category,) in cursor.execute(
                        "SELECT category_name FROM category_to_data WHERE scraped_data_url=?;",
                        [url],
                    )
                ]

                yield ScrapedData(
                    url=url,
                    last_scraped=last_scraped,
                    title=title,
                    content=content,
                    author=author,
                    date=date,
                    categories=categories,
                )

    def filter_scraped_urls(self, urls: List[str]) -> List[str]:

        # connect to db and check for each url if such url was scraped,
        # checking its last_scrape field

        res = []
        with sqlite3.connect(self._db_path) as connection:

            cursor = connection.cursor()
            for url in urls:

                # if last_scraped is null, then it wasn't scraped
                is_there = cursor.execute(
                    "SELECT 1 FROM scraped_data WHERE url=? AND last_scraped IS NOT NULL",
                    [url],
                ).fetchone()

                if not is_there:
                    res.append(url)

        return res

    def was_scraped(self, url: str) -> bool:

        # Connect to db and check if object was scraped
        with sqlite3.connect(self._db_path) as connection:
            cursor = connection.cursor()

            obj = cursor.execute(
                "SELECT ? FROM scraped_data WHERE last_scraped IS NOT NULL", [url]
            ).fetchone()

            return obj != None

    def save(self, url_data: List[ScrapedData]):
        # Save data into database
        if not url_data:
            return

        with sqlite3.connect(self._db_path) as connection:
            cursor = connection.cursor()
            data_to_insert = [dataclasses.asdict(data) for data in url_data]
            cursor.executemany(
                "INSERT OR REPLACE INTO scraped_data VALUES (:url, :last_scraped, :title, :content, :author, :date)",
                data_to_insert,
            )

            for data in url_data:
                # insert new categories
                # may have none if category field is set to None
                if data.categories is None: continue

                cursor.executemany(
                    "INSERT OR IGNORE INTO category VALUES (?)",
                    [(cat,) for cat in data.categories],
                )

                # insert many to many relationship
                cursor.executemany(
                    "INSERT OR IGNORE INTO category_to_data VALUES (?, ?)",
                    [(data.url, category) for category in data.categories],
                )

            # save changes
            connection.commit()

    def delete(self, urls: List[str]):

        # Bulk delete provided urls
        with sqlite3.connect(self._db_path) as connection:
            cursor = connection.cursor()
            cursor.executemany(
                "DELETE FROM scraped_data WHERE url=?", [(url,) for url in urls]
            )
            connection.commit()


def _parse_dict_to_url_data(obj: Dict[str, Any]) -> ScrapedData:
    """
    Parse dict object into a new object instance
    """
    return ScrapedData(**obj)



# Python imports
from dataclasses import dataclass, asdict
from typing import List, Dict, Any
from datetime import datetime
import json


@dataclass()
class ScrapedData:
    """
        This is a general data format class, 
        every data format for other scrapers could have 
        additional fields according to its needs and
        scrapable data, but then they should be able to 
        convert themselves into this format, possibly leaving a 
        few fields as None. Thus, we can be able to 
        easily map from a scrapers's output to a database 
        scheme
    """

    url: str
    last_scraped: datetime = None
    title: str = None
    content: str = None
    author: str = None
    categories: List[str] = None
    date: str = None
      
    def pretty_print(self, max_content_len : int = -1) -> str:
        """
            Return a human-readable representation of this data.
            Truncate content if requested
        """

        # create categories string 
        categories = "".join(map(lambda s: f"\t+ {s}\n", self.categories))

        max_content_len = max_content_len if max_content_len > 0 else len(self.content)

        # create body content:
        content = self.content
        if max_content_len < len(self.content):
            content = content[:max_content_len] + "..."
        
        return f"title: {self.title}\nauthor: {self.author}\ndate: {self.date}\ncategories:\n{categories}content:\n\t{content}"

class ScrapedDataEncoder(json.JSONEncoder):
    """
        Encoder to turn this file into json format
    """

    def default(self, obj: ScrapedData) -> Dict[str, Any]:
        if isinstance(obj, ScrapedData):
            return asdict(obj)
        elif isinstance(obj, datetime):

            # Local imports
            from config import settings

            return datetime.strftime(obj, settings.date_format)

        return json.JSONEncoder.default(self, obj)



"""
    Crawler for ElPitazo site. 
"""

# Local imports
from c4v.scraper.crawler.crawlers import base_crawler

# Third party imports


# Python imports
from typing import List


class ElPitazoCrawler(base_crawler.BaseCrawler):
    """
        Class to crawl El Pitazo urls 
    """

    start_sitemap_url = "https://elpitazo.net/sitemap.xml"
    name = "el_pitazo"
    
    @staticmethod
    def should_crawl(url: str) -> bool:
        return url.startswith("https://elpitazo.net/sitemap-pt-post-")



"""
    Base class for url crawlers. Based on sitemap crawling.
"""
# Python imports
from typing import Any, List, Callable
import sys

# Third party imports
import requests
import bs4


class BaseCrawler:
    """
        Inherit this class to create a new crawler.
        Probably, the only function you might want to implement is check_sitemap_url,
        that checks if an url in the sitemap index corresponds to an interesting subset 
        of pages

        You might also want to override should_scrape method if you want to add detailed 
        filtering to over urls to be retrieved 
    """

    start_sitemap_url: str = None  # Override this field to define sitemap to crawl
    name: str = None               # Crawler name, required to identify this crawler 

    def crawl_urls(self, up_to : int = None) -> List[str]:
        """
            Return a list of urls scraped from the site intended for this scraper.
            Parameters:
                up_to : int = Maximum amount of elements to store. Retrieve all if no number is provided, should be positive 
            Return:
                List of urls from sitemap
        """

        # Set up max size
        up_to = up_to or sys.maxsize

        # Check for consistency
        if up_to <= 0:
            raise ValueError("Max size should be a possitive number")

        # Set up storing function 
        items = []
        def store_items(new_items : List[str]):
            rem = up_to - len(items)
            items.extend(new_items[:rem])

        # Set up stop function 
        def should_stop_when() -> bool:
            return len(items) >= up_to

        # crawl for urls
        self.crawl_and_process_urls(store_items, should_stop_when)

        return items

    def crawl_and_process_urls(
        self, 
        post_process_data: Callable[[List[str]], Any] = None,
        should_stop: Callable[[], bool] = None
    ):
        """
            crawl urls, processing them with the provided function
            Parameters:
                + post_process_data : ([str]) -> [str] = function to call over the resulting set of urls. May be called in batches
                + should_stop : () -> bool = function to every iteration to check if the crawling process should stop
            Return:
                List of urls from sitemap
        """

        # Url accumulator
        urls: List[str] = []

        # sitemaps to parse
        sitemaps = self.get_sitemaps_from_index()

        # Set up stop function
        should_stop = should_stop or (lambda _: False)

        for sitemap in sitemaps:

            # Get a sitemap from its url
            sitemap_content = self.get_sitemap_from_url(sitemap)

            # parse urls for the current sitemap
            new_urls = self.parse_urls_from_sitemap(sitemap_content)
            urls.extend(new_urls)

            # process new urls if post process function exists
            if post_process_data:
                post_process_data(new_urls)

            if should_stop():
                break

    def get_sitemaps_from_index(self) -> List[str]:
        """
            Some sites may have its sitemap paginated in more sitemaps.
            I such case, override this class. By default, it assumes 
            the sitemap set is actually the starting one.
        """
        assert self.start_sitemap_url != None, "Start sitemap url not configured"

        resp = requests.get(self.start_sitemap_url)
        if resp.status_code != 200:
            resp.raise_for_status()

        return self.parse_sitemaps_urls_from_index(resp.text)

    def parse_sitemaps_urls_from_index(self, sitemap_index: str) -> List[str]:
        """
            Get sitemap list from the content of an xml file with the filemap data
            Parameters:
                + sitemap_index : str = sitemap index xml content as string
            Return:
                List of urls listed by this index
        """
        soup = bs4.BeautifulSoup(sitemap_index, "xml")
        urls = map(lambda l: l.get_text(), soup.find_all("loc"))
        urls = filter(self.should_crawl, urls)

        return list(urls)

    def get_sitemap_from_url(self, sitemap_url: str) -> str:
        """
            Get sitemap xml content from its corresponding url
            Parameters:
                + sitemap_url : str = sitemap url to retrieve
            Return:
                sitemap content if everything went ok
        """
        resp = requests.get(sitemap_url)
        if resp.status_code != 200:
            resp.raise_for_status()

        return resp.text

    def parse_urls_from_sitemap(self, sitemap: str) -> List[str]:
        """
            Given a sitemap body, parse site urls from it 
            Parameters:
                + sitemap : str = sitemap xml content as string
            Return:
                list of urls parsed for this sitemap
        """
        soup = bs4.BeautifulSoup(sitemap, "xml")
        urls = map(lambda l: l.get_text(), soup.select("url > loc"))
        urls = filter(self.should_scrape, urls)

        return list(urls)

    @staticmethod
    def should_crawl(url: str) -> bool:
        """
            Function to check if a given sitemap url
            to another sitemap is a desired one
            Parameters:
                + url : str = url to check
            return:
                if this is a valid url
        """
        raise NotImplementedError("Implement this abstract method")

    @staticmethod
    def should_scrape(url: str) -> bool:
        """
            Function to check if an url to a web page in the site is a valid one
            Parameters:
                + url : str = a web page url to check
            Return:
                boolean, telling if this is a valid url
        """
        return True



"""
    This classifier class will perform basic experiments, receiving 
    as arguments the training arguments and the columns to use from the training dataset
"""
# Python imports
from typing             import Dict, List, Any, Tuple
from pathlib            import Path
from importlib          import resources
from enum               import Enum
# Third Party
from pandas.core.frame  import DataFrame
from transformers import (
    RobertaTokenizer,
    Trainer,
    TrainingArguments,
    RobertaForSequenceClassification,
)
from sklearn.metrics import accuracy_score, f1_score, recall_score, precision_score
from sklearn.model_selection import train_test_split
from datasets import Dataset
import torch
import pandas as pd
import numpy as np
import os

from transformers.trainer_utils import EvalPrediction

device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

BASE_C4V_FOLDER = os.path.join(os.environ.get("HOME"), "/.c4v")
BASE_C4V_EXPERIMENTS_FOLDER = os.path.join(BASE_C4V_FOLDER, "/experiments")

class Tags(Enum):
    """
        Possible tags variants for an article
    """
    DENUNCIA_FALTA_DEL_SERVICIO = "DENUNCIA FALTA DEL SERVICIO"

class ClassifierExperiment:
    """
        This class provides a simple way to run simple experiments.
    """

    LOGS_FOLDER_NAME: str = "logs"
    RESULTS_EXPERIMENT_NAME: str = "results"
    RELEVANT_LABEL: str = "DENUNCIA FALTA DEL SERVICIO"

    def __init__(
        self,
        branch_name: str,
        experiment_name: str,
        test_dataset: str = "elpitazo_positivelabels_devdataset.csv",
        columns: List[str] = ["text"],
        base_path: str = BASE_C4V_EXPERIMENTS_FOLDER,
        use_cuda: bool = True,
        model_name: str = "mrm8488/RuPERTa-base",
        train_args: TrainingArguments = None,
    ):
        self._columns = columns
        self._branch_name = branch_name
        self._experiment_name = experiment_name
        self._experiments_folder = base_path
        self._device = (
            torch.device("cuda")
            if use_cuda and torch.cuda.is_available()
            else torch.device("cpu")
        )
        self._test_dataset = test_dataset
        self._model_name = model_name
        self._train_args = train_args

    def _get_files_path(self) -> str:
        """
            Get path to files for this experiment
        """
        return os.path.join(
            self._experiments_folder, f"{self._branch_name}/{self._experiment_name}"
        )

    def _get_path_to(self, folder: str) -> str:
        """
            Get path to given folder and create it if doesn't exist
        """
        if not os.path.exists(folder):
            Path(folder).mkdir(parents=True)

        return folder

    def get_logs_path(self) -> str:
        """
            Get path to logs for this experiment
            Return:
                path to log folder where we store logs for this experiment
        """
        return self._get_path_to(
            os.path.join(self._get_files_path(), f"{self.LOGS_FOLDER_NAME}")
        )

    def get_results_path(self) -> str:
        """
            Get path to results for this experiment
            Return:
                path to results folder where we store results for this experiment
        """
        return self._get_path_to(
            os.path.join(self._get_files_path(), f"{self.RESULTS_EXPERIMENT_NAME}")
        )

    def get_experiments_path(self) -> str:
        """
            Get path to experiments folder
        """
        return self._get_path_to(self._get_files_path())

    def get_dataframe(self, dataset_name: str = None) -> DataFrame:
        """
            Get dataframe as a pandas dataframe 
        """
        with resources.open_text(
            "data.raw.huggingface", dataset_name or self._test_dataset
        ) as f:
            return pd.read_csv(f)

    def prepare_dataframe(self) -> Tuple[List[str], List[int]]:
        """
            Return the list of text bodies and its corresponding label of whether it is 
            a missing service problem or not, expressed as int
            Return:
                ([str], [int]) = the i'st position of the first list is the body of a news article, and the 
                                 i'st position of the second list tells whether the article i talks about
                                 a missing service or not, expressed as an int (1 if it is, 0 if not)
        """

        df_elpitazo_pscdd = self.get_dataframe()
        df_elpitazo_pscdd["label"] = (
            df_elpitazo_pscdd.tipo_de_evento == Tags.DENUNCIA_FALTA_DEL_SERVICIO.value
        ).astype(int)

        df_elpitazo_pscdd = df_elpitazo_pscdd.convert_dtypes()
        df_issue_text = df_elpitazo_pscdd[[*self._columns, "label"]]
        df_issue_text.dropna(inplace=True)

        if len(self._columns) == 1:
            x = list(df_issue_text["text"])
        else:
            x = list(zip(*[list(df_issue_text[col]) for col in self._columns]))

        y = list(df_issue_text["label"])

        return x, y

    def load_model_from_hub(self) -> RobertaForSequenceClassification:
        """
            Return the model from hub, already configured
            Return:
                Model itself, already configured and loaded to device
        """
        model = RobertaForSequenceClassification.from_pretrained(self._model_name, num_labels=2)
        # send model to device
        model.to(self._device)

        return model

    def load_tokenizer_from_hub(self) -> RobertaTokenizer:
        """
            Load and configure tokenizer from hub
            Return:
                tokenizer from hub
        """
        tokenizer = RobertaTokenizer.from_pretrained(self._model_name)

        return tokenizer

    def transform_dataset(
        self, x: List[str], y: List[int], tokenizer: RobertaTokenizer
    ) -> Tuple[Dataset, Dataset]:
        """
            perform operations needed to post process a dataset separated in input list
            "x" and expected answers list "y", returning a training and a validation datasets,
            ready to be fed into the model 
            Parameters:
                x : [str] = List of content inputs to train with
                y : [int] = List of integer values matching answer for inputs in x list
            Return:
                (Dataset, Dataset) = the training and the validation dataset, in such order
        """
        # Train Test Split
        X_train, X_val, y_train, y_val = train_test_split(x, y, test_size=0.2)

        X_train_tokenized = tokenizer(
            X_train, padding=True, truncation=True, max_length=512
        )

        X_val_tokenized = tokenizer(
            X_val, padding=True, truncation=True, max_length=512
        )

        # Create torch dataset
        class _Dataset(torch.utils.data.Dataset):
            def __init__(self, encodings, labels=None):
                self.encodings = encodings
                self.labels = labels

            def __getitem__(self, idx):
                item = {
                    key: torch.tensor(val[idx]) for key, val in self.encodings.items()
                }
                if self.labels:
                    item["labels"] = torch.tensor(self.labels[idx])
                return item

            def __len__(self):
                return len(self.encodings["input_ids"])

        train_dataset = _Dataset(X_train_tokenized, y_train)
        val_dataset = _Dataset(X_val_tokenized, y_val)

        return train_dataset, val_dataset

    @staticmethod
    def compute_metrics(predictions: EvalPrediction) -> Dict[str, Any]:
        """
            Use this function to evaluate metrics for a given prediction 
            Parameters:
                predictions : EvalPrediction = Prediction provided by the model with certain input 
            Return:
                Dict with computed metrics. Provided fields:
                    - accuracy
                    - precision
                    - recall
                    - f1
        """
        pred, labels = predictions
        pred = np.argmax(pred, axis=1)

        accuracy = accuracy_score(y_true=labels, y_pred=pred)
        recall = recall_score(y_true=labels, y_pred=pred)
        precision = precision_score(y_true=labels, y_pred=pred)
        f1 = f1_score(y_true=labels, y_pred=pred)

        return {
            "accuracy": accuracy,
            "precision": precision,
            "recall": recall,
            "f1": f1,
        }

    def _get_default_train_args(self) -> Dict[str, Any]:
        """
            Return a default version of training arguments as a dict
        """

        default = {
            "output_dir": self.get_results_path(),
            "num_train_epochs": 1,
            "per_device_train_batch_size": 10,
            "per_device_eval_batch_size": 10,
            "warmup_steps": 500,
            "weight_decay": 0.01,
            "logging_dir": self.get_logs_path(),
        }

        return default

    def _override_train_args(self, new_args: Dict[str, Any] = {}) -> Dict[str, Any]:
        """
            Return the default args dict, with overriden settings specified as the ones specified in 
            input dict
        """

        default = self._get_default_train_args()
        for (k, v) in new_args.items():
            default[k] = v

        return default

    def train_and_save_model(
        self,
        model: Any,
        output_dir: str = None,
        logging_dir: str = None,
        path_to_save_checkpoint: str = None,
        train_args: Dict[str, Any] = None,
        train_dataset: Dataset = None,
        eval_dataset: Dataset = None,
    ) -> Trainer:
        """
            Train and save using provided model, using provided args and storing 
            results to "output_dir" and logging additional info to "logging_dir",
            saving checkpoint to "path_to_save_checkpoint". If some of this values is not provided,
            it will be defaulted to this experiment's dedicated folder 
            Parameters:
                model = model to train
                output_dir : str = results folder, defaulted to experiment's results folder
                logging_dir : str = logs folder, defaulted to experiment's logs folder
                path_to_save_checkpoint : str = path to save checkpoint after training is finished, defaults to experiment's folder
                train_args : TrainArguments = if provided, override all default parameters passed to the training
            Return:
                properly configured trainer instance

        """
        if output_dir:
            train_args["output_dir"] = output_dir
        if logging_dir:
            train_args["logging_dir"] = logging_dir

        args = TrainingArguments(**self._override_train_args(train_args))

        trainer = Trainer(
            model=model,
            args=args,
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            compute_metrics=self.__class__.compute_metrics,
        )

        # Train pre-trained model
        trainer.train()

        model.save_pretrained(path_to_save_checkpoint or self._get_files_path())

        return trainer

    def load_fine_tuned_model(
        self, path: str = None
    ) -> RobertaForSequenceClassification:
        """
            load fine tuned model from provided path, defaults to an already trained one 
            in experiment's folder
        """
        path = path or os.path.join(self.get_experiments_path())
        model_path = os.path.join(path, "pytorch_model.bin")
        model = RobertaForSequenceClassification.from_pretrained(model_path)
        return model

    def evaluate_metrics(self, trainer: Trainer, val_dataset: Dataset) -> DataFrame:
        """
            Compute metrics as a dataframe for this trainer object using the validation dataset
            Parameters:
                trainer : Trainer = trainer object already trained and ready to test
                val_dataset : Dataset = data used for validation
            Return:
                A dataframe with output of a validation
        """
        ## Evaluate Metrics
        metrics = trainer.evaluate(val_dataset)
        metrics_df = pd.DataFrame.from_dict(
            metrics, orient="index", columns=["metrics_value"]
        )
        return metrics_df

    def run_experiment(self, train_args: Dict[str, Any] = None):
        """
            Run an experiment using the give training args
        """
        # Prepare dataframe and load model + tokenizer
        x, y = self.prepare_dataframe()

        model = self.load_model_from_hub()
        tokenizer = self.load_tokenizer_from_hub()

        train_dataset, val_dataset = self.transform_dataset(x, y, tokenizer)

        # Fine tune the model
        fine_tuned_model_trainer = self.train_and_save_model(
            model=model,
            output_dir=self.get_results_path(),
            logging_dir=self.get_logs_path(),
            path_to_save_checkpoint=self.get_experiments_path(),
            train_args=self._override_train_args(train_args or {}),
            train_dataset=train_dataset,
            eval_dataset=val_dataset,
        )

        # Get the metrics from the model
        metrics_df = self.evaluate_metrics(
            trainer=fine_tuned_model_trainer, val_dataset=val_dataset
        )

        return metrics_df




import logging
from google.cloud import bigquery
from google.oauth2 import service_account
import os

logger = logging.getLogger(__name__)

SERVICE_ACCOUNT_KEY_PATH_ENV_VAR = "ANGOSTURA_SERVICE_ACCOUNT"


class AngosturaLoader:
    def __init__(self, service_account_key_path=None):
        """
        Creates connection to perform queries to Code For Venezuela's Angostura ETL.
        BigQuery is used as warehouse.
        
        Parameters:
        ------

            service_account_key_path: String, Default = None.
                Path to service account json key. 
                This json key is created by the developers of Angostura, to have 
                access to this key please contact the C4V team.
        
        Attributes:
        ----------

            client: google.cloud.bigquery.client.Client object.
                Connection object to Angostura.

            key: String.
                Path to service account json key. 

        Examples:
        ---------
        
            # Note - It is necessary to have the key angostura_connection.json in the 
            # root directory. If you don't have it please contact anyone from c4v's Sambil team.
            test = AngosturaLoader()
            
            df = test.create_query(
                    "SELECT * FROM `event-pipeline.angostura.sinluz_rawtweets` LIMIT 1"
            )
            
            print(df.head())
        """
        self.client = None

        if service_account_key_path:
            logger.info("Using user input key path.")
            self.key_path = service_account_key_path
        elif SERVICE_ACCOUNT_KEY_PATH_ENV_VAR in os.environ:
            logger.info("Using environment variable.")
            self.key_path = os.environ[SERVICE_ACCOUNT_KEY_PATH_ENV_VAR]
        else:
            logger.info("Looking for key in project's root directory.")
            self.key_path = "angostura_connection.json"

    def create_connection(self):
        """
        Creates connection with angostura.
        
        Parameters:
        ----------

            Self: Object.
                Object instance.
        
        Returns:
        --------

            self.client: google.cloud.bigquery.client.Client object
                Connection object to Angostura.
        """

        credentials = service_account.Credentials.from_service_account_file(
            self.key_path, scopes=["https://www.googleapis.com/auth/cloud-platform"],
        )

        self.client = bigquery.Client(
            credentials=credentials, project=credentials.project_id,
        )

    def create_query(self, query):
        """
        Query Angostura's warehouse.
        
        Parameters:
        ----------

            Query: String.
                SQL query with BigQuery syntax.

        Returns:
        --------
            results_df: Pandas.DataFrame n,n.
                Pandas Dataframe with query results.
        """

        if self.client is None:
            self.create_connection()

        logger.info("Querying Database...")
        QUERY = query

        results_df = self.client.query(QUERY).to_dataframe()

        return results_df



"""
    For now, this is going to be the main CLI executable,
    so we can test things in the meanwhile
"""
# Third party imports
import click

# Python imports
from typing import List
import os

# Local imports
from c4v.scraper.scraper import bulk_scrape
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData
from c4v.scraper.settings import INSTALLED_CRAWLERS
from c4v.scraper.persistency_manager.sqlite_storage_manager import SqliteManager
from c4v.scraper.utils import data_list_to_table_str
from c4v.scraper.scraper    import bulk_scrape
from c4v.scraper.settings   import INSTALLED_CRAWLERS

# Folder to search for local files @TODO debo cambiar esto para escoger correctamente el sitio para el directorio
DEFAULT_FILES_FOLDER = os.environ.get("HOME") + "/.c4v"
DEFAULT_DB = os.path.join(DEFAULT_FILES_FOLDER, "c4v_db.sqlite")


@click.group()
def c4v_cli():
    """
        Command entry point 
    """
    # init files if necessary:
    if not os.path.isdir(DEFAULT_FILES_FOLDER):
        click.echo(
            f"[INFO] Creating local files folder at: {DEFAULT_FILES_FOLDER}", err=True
        )
        try:
            os.mkdir(DEFAULT_FILES_FOLDER)
        except Exception as e:
            print(e)


@c4v_cli.command()
@click.option(
    "--files",
    is_flag=True,
    help="Interpret url list as files instead of urls, so urls are retrieved from such files. The file is expected to be formatted such that there's one url per line",
)
@click.option(
    "--loud", is_flag=True, help="Print scraped data to stdio as it is being scraped"
)
@click.option(
    "--up-to",
    default=-1,
    help="Maximum amount of elements to scrape. Scrape all if no argument given",
)
@click.argument("urls", nargs=-1)
def scrape(
    urls: List[str] = None, files: bool = None, loud: bool = False, up_to: int = -1
):
    """
        Use this command to run a scraping process. Scraped urls are retrieved from database by default.\n
        If provided with a list of urls, then those urls are the ones to be scraped and stored. \n
        You can pass the --files flag to interpret those paths as paths to files storing the actual urls,\n
        formated as one url per line.\n
        Parameters:\n
            + urls : [str] = List of urls or files to get urls from\n
            + files : bool = if urls are in files or not, defaults to false\n
            + up_to : int = maximum amount of urls to scrape in this pass. Scrape all non-scraped if not provided\n
            + loud : bool = if should print scraped data once a scraping is finished\n
    """

    db_manager = SqliteManager(DEFAULT_DB)

    # Read urls
    urls_to_scrape = []

    if not urls:
        urls_to_scrape = [d.url for d in db_manager.get_all(up_to, scraped=False)]
    elif files:  # if urls are stored in files
        for file in urls:  # iterate over every file
            try:
                with open(file) as urls_file:
                    content = map(
                        lambda s: s.strip(), urls_file.readlines()
                    )  # parse every line as a single url
                    urls_to_scrape.extend(content)
            except IOError as e:
                click.echo(f"Could not open input file: {file}. Error: {e.strerror}")
    else:
        urls_to_scrape = urls

    # scrape urls
    scraped_data = bulk_scrape(urls_to_scrape)

    # Write output
    db_manager.save(scraped_data)

    # Print obtained results if requested
    if loud:
        click.echo(data_list_to_table_str(scraped_data))


@c4v_cli.command()
@click.option("--list", is_flag=True, help="list available crawlers")
@click.option("--all", is_flag=True, help="Run all available crawlers")
@click.option("--all-but", is_flag=True, help="Run all crawlers except listed ones")
@click.option("--loud", is_flag=True, help="Print results to terminal")
@click.argument("crawlers", nargs=-1)
def crawl(
    crawlers: List[str] = [],
    list: bool = False,
    all: bool = False,
    all_but: bool = False,
    loud: bool = False,
):
    """
        Crawl for new urls, ignoring already scraped ones.\n
        Parameters:\n
            + list : bool = if should print list of available crawlers by name\n
            + all  : bool = if should crawl all crawlers. Not true by default as a crawling process may be really slow\n
            + all_but : bool = run all crawlers except the ones provided as arguments\n
            + loud : bool = if should print scraped urls to stdio\n
    """
    crawlable_sites = "".join([f"\t{crawl.name}\n" for crawl in INSTALLED_CRAWLERS])

    # Create default db manager
    db_manager = SqliteManager(DEFAULT_DB)

    # list available crawlers if requested to
    if list:
        click.echo(f"Crawlable sites: \n{crawlable_sites}")
        if not crawlers and not all:  # if nothing else to do, just end
            return

    # Check for errors:
    # if no crawlers to use, just end
    if not crawlers and not all:
        click.echo(f"Not crawlers provided. Available crawlers:\n{crawlable_sites}")
        return

    # raise a warnning if incompatible flags where provided
    if all_but and all:
        click.echo(
            f"[WARNING] --all and --all-but incompatible flags were provided, using only all"
        )

    # set up crawlers to run
    if all:
        crawlers_to_run = INSTALLED_CRAWLERS
    elif all_but:
        crawlers_to_run = [
            crawler for crawler in INSTALLED_CRAWLERS if crawler.name not in crawlers
        ]
    else:
        crawlers_to_run = [
            crawler for crawler in INSTALLED_CRAWLERS if crawler.name in crawlers
        ]

    # function to format crawled urls list
    format_url_list = lambda list: "".join([f"{s}\n" for s in list])

    # if loud flag is provided, print it tu stdio
    for crawler in crawlers_to_run:
        c = crawler()

        def process(list: List[str]):
            scraped_data = [
                ScrapedData(url=url) for url in db_manager.filter_scraped_urls(list)
            ]
            db_manager.save(scraped_data)

            if loud:
                click.echo(format_url_list(list))

        c.crawl_and_process_urls(process)


@c4v_cli.command()
@click.option("--urls", is_flag=True, help="Only list urls")
@click.option("--limit", default=100, help='List only up to "limit" rows')
@click.option("--col-len", default=50, help="Columns max length")
@click.option(
    "--scraped-only",
    default=None,
    help="Retrieve only complete rows, those with its scraped data",
)
def list(
    urls: bool = False, limit: int = 100, col_len: int = 50, scraped_only: bool = None
):
    """
    List requested info as specified by arguments.\n
    Parameters:\n
        + urls : bool = list urls only\n
        + limit : int = max num of rows to show. Set -1 to list all\n
        + col_len : int = Maximum lenght of shown columns\n
        + scraped_only : bool = if true, retrieve already scraped ones only, if false, retrieve non scraped only. If not provided, return anything\n
    """

    scraped_only = (
        scraped_only == "true" or scraped_only == "True" or scraped_only == "1"
    )

    db_manager = SqliteManager(DEFAULT_DB)

    # Just print urls if requested so
    if urls:
        for data in db_manager.get_all(limit):
            click.echo(data.url)
        return

    # Get printable version of retrieved data
    data_to_print = data_list_to_table_str(
        [d for d in db_manager.get_all(limit, scraped_only)], max_cell_len=col_len
    )

    click.echo(data_to_print)
    print(scraped_only)

if __name__ == "__main__":
    c4v_cli()



"""
    Base Persistency Manager: We need a way to handle persistency, 
    as simple files or maybe with a more complex scheme, such as databases
    or http requests. This base class provides a contract that
    every persistency manager should match in order to be used with our app.
"""

# Python imports
from typing import Iterator, List

# Local imports
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData


class BasePersistencyManager:
    """
        Base class to provide support for persistency management
    """

    def get_all(self, limit : int, scraped : bool) -> Iterator[ScrapedData]:
        """
            Return an iterator over the set of stored instances
            Parameters:
                + limit : int = Max amount of elements to retrieve
                + scraped : bool = True if retrieved data should be scraped, false if it shouldn't, None if not relevant
            Return:
                Iterator of stored ScrapedData instances
        """
        raise NotImplementedError("Implement filter_scraped_urls abstract function")

    def filter_scraped_urls(self, urls: List[str]) -> List[str]:
        """
            Filter out urls whose data is already known, leaving only the ones to be scraped
            for first time
            Parameters: 
                urls : [str] = List of urls to filter 

            Return:
                A list of urls such that none of them has been scraped 
        """
        raise NotImplementedError("Implement filter_scraped_urls abstract function")

    def was_scraped(self, url: str) -> bool:
        """
            Tells if a given url is already scraped (it's related data is already know)
            Parameters:
                url : str = url to check if it was already scraped
            Return:
                If the given url's related data is already known
                
        """
        raise NotImplementedError("Implement was_scraped abstract function")

    def save(self, url_data: List[ScrapedData]):
        """
            Save provided data to local storage. 
            If some some urls are already in local storage, update them with provided new data.
            If not, delete them.
            Parameters:
                - data : [ScrapedData] = data to be saved
        """
        raise NotImplementedError("Implement save abstract method")

    def delete(self, urls: List[str]):
        """
            Delete provided urls from persistent storage
            Parameters:
                - urls : [str] = Urls to be deteled
        """
        raise NotImplementedError("Implement delete abstract method")



"""
    Base class for scraped data
"""

# Local imports
from .scraped_data import ScrapedData

# Python imports:
from dataclasses    import dataclass
from datetime       import datetime

@dataclass(frozen=True)  # Readonly attributes
class BaseDataFormat:
    """
        Base class for every scraper output class
    """
    url : str
    last_scraped : datetime
    
    def to_scraped_data(self) -> ScrapedData:
        """
            This function transforms from a data format 
            to our standard ScrapedData format
        """
        raise NotImplementedError(
            "Every BaseScrapedData sub class should implement to_scraped_data method"
        )



"""
    This file manages multiple settings shared across the scraper,
    such as mappings from urls to scrapers
"""
from c4v.scraper.crawler.crawlers.base_crawler import BaseCrawler
from c4v.scraper.scrapers.base_scraper import BaseScraper
from c4v.scraper.scrapers.el_pitazo_scraper import ElPitazoScraper
from c4v.scraper.scrapers.primicia_scraper import PrimiciaScraper
from c4v.scraper.crawler.crawlers import primicia_crawler, el_pitazo_crawler
from c4v.scraper.utils import check_scrapers_consistency

import os
from typing import Type, List


INSTALLED_SCRAPERS: List[Type[BaseScraper]] = [ElPitazoScraper, PrimiciaScraper]

INSTALLED_CRAWLERS: List[Type[BaseCrawler]] = [primicia_crawler.PrimiciaCrawler, el_pitazo_crawler.ElPitazoCrawler]

SUPPORTED_DOMAINS  = [s.intended_domain for s in INSTALLED_SCRAPERS]

# Check for scraper consistency
check_scrapers_consistency(INSTALLED_SCRAPERS)

# Dict with information to map from domain to
# Spider
URL_TO_SCRAPER = {s.intended_domain: s for s in INSTALLED_SCRAPERS}

# root dir, so we can get resources from module directories
ROOT_DIR = os.path.dirname(os.path.abspath(__file__))


"""
    Base class for ElPitazo site scraped data. Includes:
        + body
        + title
        + author
        + tags
        + categories
"""
# Local imports
from c4v.scraper.scraped_data_classes.base_scraped_data import BaseDataFormat
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData

# Python imports
from dataclasses import dataclass
from typing import List


@dataclass(frozen=True)
class ElPitazoData(BaseDataFormat):
    """
        Scraped data expected for ElPitazo news articles
    """

    tags: List[str]
    categories: List[str]
    title: str
    author: str
    date: str
    body: str

    def to_scraped_data(self) -> ScrapedData:
        return ScrapedData(
            author=self.author,
            date=self.date,
            title=self.title,
            categories=self.tags + self.categories,
            content=self.body,
            url= self.url,
            last_scraped=self.last_scraped
        )



"""
    Main module interface
"""

# Local imports
from c4v.scraper.scraped_data_classes.base_scraped_data import BaseDataFormat
from c4v.scraper.scrapers.base_scraper import BaseScraper
from c4v.scraper.scraped_data_classes.scraped_data import ScrapedData
from .settings import URL_TO_SCRAPER
from c4v.scraper.utils import get_domain_from_url, valid_url

# Python imports
from typing import List, Type, Dict, Any


def scrape(url: str) -> ScrapedData:
    """
        Scrape data for the given url if such url is scrappable,
        Raise ValueError if not. 

        Parameters:
            + url - str : Url to scrape
        Return:
            A ScrapedData object describing the data that could be 
            extracted for this url. Obtained data depends on the url itself, 
            so available data may change depending on the scrapped url, some fields 
            may be null.
    """
    scraper = _get_scraper_from_url(url)()
    return scraper.scrape(url)


def bulk_scrape(urls: List[str]) -> List[ScrapedData]:
    """
        Performs a bulk scraping over a list of urls.
        Order in the item list it's not guaranteed to be
        the same as in the input list

        Parameters:
            + urls : [str] = Urls to be scraped
        Return:
            A list of items scraped for each url in the original list
    """

    scrapers: Dict[Type[BaseScraper], List[str]] = {}

    # Classify urls to its according scraper
    for url in urls:
        scraper = _get_scraper_from_url(url)

        url_list = scrapers.get(scraper)
        if not (url_list):
            url_list = scrapers[scraper] = []

        url_list.append(url)

    # Schedule scraping
    scrapers_instances: List[BaseScraper] = []
    for (scraper, url_list) in scrapers.items():
        s = scraper()  # Create a new scraper instance
        s.schedule_scraping(url_list)
        scrapers_instances.append(s)

    # TODO: write a cleaner and transparent interface to show that scrapy scrapers only blocks once

    # start scraping for every scraper
    for s in scrapers_instances:
        s.start_bulk_scrape()

    # Retrieve scraped items
    items: List[BaseDataFormat] = []
    for s in scrapers_instances:
        items.extend(s.get_scraped_items())

    for i in range(len(items)):
        items[i] = items[i].to_scraped_data()
    return items


def _get_scraper_from_url(url: str) -> Type[BaseScraper]:
    """
        Validates if this url is scrapable and returns its 
        corresponding spider when it is.
        Raise ValueError if url is not valid or if it's not 
        scrapable for our supported scrapers
    """

    if not valid_url(url):
        raise ValueError(f"This is not a valid url: {url}")

    domain = get_domain_from_url(url)
    scraper = URL_TO_SCRAPER.get(domain)
    if not (scraper):
        raise ValueError(f"Unable to scrap this url: {url}")

    return scraper
